import itertools

import numpy as np

from stimupy.components import combine_masks, draw_regions
from stimupy.components.shapes import rectangle
from stimupy.stimuli import mask_targets
from stimupy.stimuli.gratings import squarewave
from stimupy.stimuli.pinwheels import pinwheel as angular
from stimupy.stimuli.waves import square_radial as radial
from stimupy.stimuli.wedding_cakes import wedding_cake

__all__ = [
    "generalized",
    "white",
    "white_two_rows",
    "anderson",
    "howe",
    "yazdanbakhsh",
    "angular",
    "radial",
    "wedding_cake",
]


def generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    intensity_bars=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    target_center_offsets=0,
    target_heights=None,
    origin="corner",
    round_phase_width=True,
):
    """General function to create White's stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    target_center_offsets : float, or Sequence[float, ...]
        center offset of targets in degrees visual angle (default: 0)
    target_heights : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1979).
        A new effect of pattern on perceived lightness.
        Perception, 8(4), 413-416.
        https://doi.org/10.1068/p080413
    """

    # Spatial square-wave grating
    stim = squarewave(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        rotation=rotation,
        phase_shift=0,
        period=period,
        intensity_bars=intensity_bars,
        origin=origin,
        round_phase_width=round_phase_width,
    )

    # Mask target bars
    target_bar_mask = mask_targets(
        element_mask=stim["grating_mask"], target_indices=target_indices
    )
    if isinstance(target_indices, (int, float)):
        target_indices = (target_indices,)
    stim["target_indices"] = target_indices

    # Mask rectangular regions
    if isinstance(target_heights, (int, float)):
        target_heights = (target_heights,)
    if isinstance(target_center_offsets, (int, float)):
        target_center_offsets = (target_center_offsets,)
    if len(target_indices) != 0 and target_heights is None:
        raise ValueError("generalized() missing argument 'target_heights' which is not 'None'")
    if len(target_indices) == 0 and target_heights is None:
        target_heights = (0,)
    target_heights = tuple(itertools.islice(itertools.cycle(target_heights), len(target_indices)))
    target_center_offsets = tuple(
        itertools.islice(itertools.cycle(target_center_offsets), len(target_indices))
    )

    stim_center = stim["visual_size"].height / 2
    target_rect_masks = []
    for target_idx, (height, offset) in enumerate(zip(target_heights, target_center_offsets)):
        # Draw a stripe of target_height x stim_width, at center + offset
        rect = rectangle(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            shape=stim["shape"],
            rectangle_size=(height, stim["visual_size"].width),
            rectangle_position=(stim_center + offset - (height / 2), 0),
            intensity_rectangle=target_idx,
        )
        target_rect_masks.append(rect["rectangle_mask"])
    stim["target_heights"] = target_heights
    stim["target_center_offsets"] = target_center_offsets

    # Combine rect & bar masks
    target_masks = []
    for target_idx, rect_mask in enumerate(target_rect_masks):
        # Find where strip intersects with the target bar
        target_mask = (target_bar_mask == target_idx + 1) & rect_mask
        target_masks.append(target_mask)

    # Combine masks
    if len(target_masks) > 0:
        target_mask = combine_masks(*target_masks)
    else:
        target_mask = np.zeros_like(stim["img"])
    stim["target_mask"] = target_mask.astype(int)

    # Draw targets
    stim["img"] = np.where(
        target_mask,
        draw_regions(mask=target_mask, intensities=intensity_target, intensity_background=0.0),
        stim["img"],
    )
    stim["intensity_target"] = intensity_target

    return stim


def white(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    intensity_bars=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    target_heights=None,
    origin="corner",
    round_phase_width=True,
):
    """White's stimulus where all targets are vertically aligned at half the stimulus height

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    target_heights : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1979).
        A new effect of pattern on perceived lightness.
        Perception, 8(4), 413-416.
        https://doi.org/10.1068/p080413
    """

    stim = generalized(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=rotation,
        intensity_bars=intensity_bars,
        target_indices=target_indices,
        intensity_target=intensity_target,
        target_center_offsets=0,
        target_heights=target_heights,
        origin=origin,
        round_phase_width=round_phase_width,
    )
    return stim


def white_two_rows(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    intensity_bars=(0.0, 1.0),
    intensity_target=0.5,
    target_indices_top=(),
    target_indices_bottom=(),
    target_center_offset=None,
    target_heights=None,
    origin="corner",
    round_phase_width=True,
):
    """White's stimulus where targets are placed in two rows (top, bottom) that have the same
    distance from the center.

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    intensity_target : float
        intensity value of target
    target_indices_top : int or tuple of ints
        bar indices where top target(s) will be placed. As many targets as ints.
    target_indices_bottom : int or tuple of ints
        bar indices where bottom target(s) will be placed. As many targets as ints.
    target_center_offset : float
        offset from target centers to image center in degree visual angle.
    target_heights : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1979).
        A new effect of pattern on perceived lightness.
        Perception, 8(4), 413-416.
        https://doi.org/10.1068/p080413
    """
    if not isinstance(target_center_offset, (float, int)):
        raise ValueError("target_center_offset should be a single float / int")
    if isinstance(target_indices_top, (float, int)):
        target_indices_top = (target_indices_top,)
    if isinstance(target_indices_bottom, (float, int)):
        target_indices_bottom = (target_indices_bottom,)

    target_indices = target_indices_top + target_indices_bottom
    offsets_top = (-target_center_offset,) * len(target_indices_top)
    offsets_bottom = (target_center_offset,) * len(target_indices_bottom)
    target_center_offsets = offsets_top + offsets_bottom

    stim = generalized(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=rotation,
        intensity_bars=intensity_bars,
        target_indices=target_indices,
        intensity_target=intensity_target,
        target_center_offsets=target_center_offsets,
        target_heights=target_heights,
        origin=origin,
        round_phase_width=round_phase_width,
    )

    stim["target_indices_top"] = target_indices_top
    stim["target_indices_bottom"] = target_indices_bottom
    return stim


def anderson(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    intensity_bars=(0.0, 1.0),
    intensity_target=0.5,
    target_indices_top=None,
    target_indices_bottom=None,
    target_center_offset=0,
    target_height=None,
    intensity_stripes=(0.0, 1.0),
    stripe_center_offset=0,
    stripe_height=None,
    round_phase_width=True,
):
    """Anderson variation of White's stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    intensity_bars : (float, float)
        intensity values of bars
    intensity_target : float
        intensity value of target
    target_indices_top : int or tuple of ints
        bar indices where top target(s) will be placed. As many targets as ints.
    target_indices_bottom : int or tuple of ints
        bar indices where bottom target(s) will be placed. As many targets as ints.
    target_center_offset : float
        offset from target centers to image center in degree visual angle.
    target_height : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    intensity_stripes : (float, float)
        intensity values of horizontal stripes
    stripe_center_offset : float
        offset from stripe centers to image center in degree visual angle.
    stripe_height = float
        stripe height in degrees visual angle
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Anderson, B. L. (2001).
        Contrasting theories of White's illusion.
        Perception, 30, 1499-1501.
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum
        of stimuli including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    """
    if target_height is None:
        raise ValueError("anderson() missing argument 'target_height' which is not 'None'")
    if stripe_height is None:
        raise ValueError("anderson() missing argument 'stripe_height' which is not 'None'")
    if isinstance(stripe_height, (int, float)):
        stripe_height = (stripe_height, stripe_height)

    # Generate White's stimulus with two rows of targets
    stim = white_two_rows(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=0.0,
        intensity_bars=intensity_bars,
        intensity_target=intensity_target,
        target_indices_top=target_indices_top,
        target_indices_bottom=target_indices_bottom,
        target_center_offset=target_center_offset,
        target_heights=target_height,
        origin="corner",
        round_phase_width=round_phase_width,
    )

    # Masks for stripes (as rectangles)
    stim_center = stim["visual_size"].height / 2
    stripe_top = rectangle(
        visual_size=stim["visual_size"],
        ppd=stim["ppd"],
        shape=stim["shape"],
        rectangle_size=(stripe_height[0], stim["visual_size"].width),
        rectangle_position=(stim_center - stripe_center_offset - (stripe_height[0] / 2), 0),
    )
    for bar_idx in stim["target_indices_top"]:
        if bar_idx < 0:
            bar_idx = int(stim["n_bars"]) + bar_idx
        stripe_top["rectangle_mask"] = np.where(
            stim["grating_mask"] == bar_idx, 0, stripe_top["rectangle_mask"]
        )

    stripe_bottom = rectangle(
        visual_size=stim["visual_size"],
        ppd=stim["ppd"],
        shape=stim["shape"],
        rectangle_size=(stripe_height[1], stim["visual_size"].width),
        rectangle_position=(stim_center + stripe_center_offset - (stripe_height[1] / 2), 0),
    )
    for bar_idx in stim["target_indices_bottom"]:
        if bar_idx < 0:
            bar_idx = int(stim["n_bars"]) + bar_idx
        stripe_bottom["rectangle_mask"] = np.where(
            stim["grating_mask"] == bar_idx, 0, stripe_bottom["rectangle_mask"]
        )

    try:
        stripes_mask = combine_masks(stripe_top["rectangle_mask"], stripe_bottom["rectangle_mask"])
    except ValueError:
        raise ValueError("Stripes overlap. Increase stripe offset or decrease stripe size.")

    # Combine images
    stripes_img = draw_regions(stripes_mask, intensities=intensity_stripes)
    img = np.where(stripes_mask, stripes_img, stim["img"])
    img = np.where(stim["target_mask"], stim["img"], img)
    stim["img"] = img

    # Output
    stim["stripe_center_offset"] = stripe_center_offset
    stim["stripe_height"] = stripe_height
    stim["stripes_mask"] = stripes_mask
    stim["intensity_stripes"] = intensity_stripes

    return stim


def howe(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    intensity_bars=(0.0, 1.0),
    intensity_target=0.5,
    target_indices_top=None,
    target_indices_bottom=None,
    target_center_offset=0,
    target_height=None,
    intensity_stripes=(0.0, 1.0),
    round_phase_width=True,
):
    """Howe variation of White's stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    intensity_bars : (float, float)
        intensity values of bars
    intensity_target : float
        intensity value of target
    target_indices_top : int or tuple of ints
        bar indices where top target(s) will be placed. As many targets as ints.
    target_indices_bottom : int or tuple of ints
        bar indices where bottom target(s) will be placed. As many targets as ints.
    target_center_offset : float
        offset from target centers to image center in degree visual angle.
    target_height : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    intensity_stripes : (float, float)
        intensity values of horizontal stripes
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum
        of stimuli including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    Howe, P. D. L. (2001).
        A comment on the Anderson (1997), the Todorovic (1997),
        and the Ross and Pessoa (2000) explanations of White's eﬀect.
        Perception, 30, 1023-1026
    """
    return anderson(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        intensity_bars=intensity_bars,
        intensity_target=intensity_target,
        target_indices_top=target_indices_top,
        target_indices_bottom=target_indices_bottom,
        target_center_offset=target_center_offset,
        target_height=target_height,
        intensity_stripes=intensity_stripes,
        stripe_center_offset=target_center_offset,
        stripe_height=target_height,
        round_phase_width=round_phase_width,
    )


def yazdanbakhsh(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    intensity_bars=(0.0, 1.0),
    intensity_target=0.5,
    target_indices_top=None,
    target_indices_bottom=None,
    target_center_offset=0,
    target_heights=None,
    # intensity_stripes=(0.0, 1.0),
    gap_size=None,
    round_phase_width=True,
):
    """Yazsdanbakhsh variation of White's stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    intensity_bars : (float, float)
        intensity values of bars
    intensity_target : float
        intensity value of target
    target_indices_top : int or tuple of ints
        bar indices where top target(s) will be placed. As many targets as ints.
    target_indices_bottom : int or tuple of ints
        bar indices where bottom target(s) will be placed. As many targets as ints.
    target_center_offset : float
        offset from target centers to image center in degree visual angle.
    target_heights : float, or Sequence[float, ...]
        height of targets in degrees visual angle
    gap_size : float
        size of gap between target and grating bar
    round_phase_width : Bool
        if True (default), round phase width of grating

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Yazdanbakhsh, A., Arabzadeh, E., Babadi, B., and Fazl, A. (2002).
        Munker-White-like illusions without T-junctions.
        Perception 31, 711-715. https://doi.org/10.1068/p3348
    """
    if target_heights is None:
        raise ValueError("yazdanbakhsh() missing argument 'target_heights' which is not 'None'")
    if gap_size is None:
        raise ValueError("yazdanbakhsh() missing argument 'gap_size' which is not 'None'")

    # Generate White's stimulus with two rows of targets
    stim = white_two_rows(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=0.0,
        intensity_bars=intensity_bars,
        intensity_target=intensity_target,
        target_indices_top=target_indices_top,
        target_indices_bottom=target_indices_bottom,
        target_center_offset=target_center_offset,
        target_heights=target_heights,
        origin="corner",
        round_phase_width=round_phase_width,
    )

    # Masks for top gaps
    stim_center = stim["visual_size"].height / 2
    gap_top_masks = []
    for t_idx, bar_idx in enumerate(stim["target_indices_top"]):
        height = stim["target_heights"][t_idx]

        # Mask for rectangle
        rect = rectangle(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            shape=stim["shape"],
            rectangle_size=(height + 2 * gap_size, stim["visual_size"].width),
            rectangle_position=(stim_center - target_center_offset - (height / 2) - gap_size, 0),
        )

        # Reduce to just this bar: intersection between rect mask and bar mask
        if bar_idx < 0:
            bar_idx = int(stim["n_bars"]) + bar_idx
        stim["gap_mask"] = (stim["grating_mask"] == bar_idx) & rect["rectangle_mask"]

        # Remove everywhere it intersects with target mask
        stim["gap_mask"] = np.where(stim["target_mask"] == t_idx + 1, 0, stim["gap_mask"])
        gap_top_masks.append(stim["gap_mask"])

    # Masks for bottom gaps
    gap_bottom_masks = []
    for idx, bar_idx in enumerate(stim["target_indices_bottom"]):
        t_idx = idx + len(stim["target_indices_top"])
        height = stim["target_heights"][t_idx]

        # Mask for rectangle
        rect = rectangle(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            shape=stim["shape"],
            rectangle_size=(height + 2 * gap_size, stim["visual_size"].width),
            rectangle_position=(stim_center + target_center_offset - (height / 2) - gap_size, 0),
        )

        # Reduce to just this bar: intersection between rect mask and bar mask
        if bar_idx < 0:
            bar_idx = int(stim["n_bars"]) + bar_idx
        stim["gap_mask"] = (stim["grating_mask"] == bar_idx) & rect["rectangle_mask"]

        # Remove everywhere it intersects with target mask
        stim["gap_mask"] = np.where(stim["target_mask"] == t_idx + 1, 0, stim["gap_mask"])
        gap_bottom_masks.append(stim["gap_mask"])

    stim["gap_mask"] = combine_masks(*gap_top_masks, *gap_bottom_masks)

    # Create stim without targets
    stim_wo = white_two_rows(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_bars=n_bars,
        bar_width=bar_width,
        period=period,
        rotation=0.0,
        intensity_bars=(intensity_bars[1], intensity_bars[0]),
        intensity_target=intensity_target,
        target_center_offset=target_center_offset,
        target_heights=target_heights,
        origin="corner",
        round_phase_width=round_phase_width,
    )

    # Remove information at gaps:
    stim["img"] = np.where(stim["gap_mask"], stim_wo["img"], stim["img"])

    # Output
    stim["gap_size"] = gap_size
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "white": white(**default_params, target_indices=(2, -3), target_heights=2, frequency=0.5,),
        "white_general": generalized(
            **default_params,
            target_indices=(1, 3, 5),
            target_center_offsets=(-1, -3, -1),
            target_heights=(2, 3, 2),
            frequency=0.5,),
        "white_two_rows": white_two_rows(
            **default_params,
            target_indices_top=(2, 4),
            target_indices_bottom=(-2, -4),
            target_heights=1,
            target_center_offset=2,
            frequency=0.5,),
        "white_anderson": anderson(
            **default_params,
            target_indices_top=3,
            target_indices_bottom=-4,
            target_center_offset=2,
            target_height=2,
            stripe_center_offset=1.5,
            stripe_height=2,
            frequency=0.5,),
        "white_howe": howe(
            **default_params,
            target_indices_top=3,
            target_indices_bottom=-4,
            target_center_offset=2,
            target_height=2,
            frequency=0.5,),
        "white_yazdanbakhsh": yazdanbakhsh(
            **default_params,
            target_indices_top=2,
            target_indices_bottom=-3,
            target_center_offset=2,
            target_heights=2,
            gap_size=0.5,
            frequency=0.5,),
        "white_angular": angular(
            **default_params,
            n_segments=10,
            target_indices=(1, 6),
            target_width=1,),
        "white_radial": radial(
            **default_params,
            frequency=0.5,
            target_indices=(1, 4),
            clip=True),
        "white_wedding_cake": wedding_cake(
            **default_params,
            L_size=(4, 2.5, 0.5),
            target_height=2,
            target_indices1=((1, 4), (1, 3)),
            target_indices2=((1, 0), (1, 1)),)
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



from .color_conversions import *
from .contrast_conversions import *
from .export import *
from .filters import *
from .masks import *
from .pad import *
from .plotting import *
from .resolution import *
from .utils import *



import numpy as np

__all__ = [
    "transparency",
    "adapt_michelson_contrast",
    "adapt_rms_contrast",
    "adapt_normalized_rms_contrast",
    "adapt_intensity_range",
    "adapt_michelson_contrast_dict",
    "adapt_rms_contrast_dict",
    "adapt_normalized_rms_contrast_dict",
    "adapt_intensity_range_dict",
]


def transparency(img, mask=None, alpha=0.5, tau=0.2):
    """Applies a transparency to image at specified (mask) location if provided

    Parameters
    ----------
    img : np.array
        image to which transparency will be applied
    mask : np.array or None (default)
        if not None, transparency will be provided at non-zero locations
        provided in this mask
    alpha : Number
        alpha of transparency (i.e. how transparent the medium is), default 0.2
    tau : Number
        tau of transparency (i.e. value of transparent medium), default 0.5

    Returns
    -------
    img with the applied transparency
    """
    if mask is None:
        mask = np.ones(img.shape)
    img = np.where(mask, alpha * img + (1 - alpha) * tau, img)
    return img


def adapt_michelson_contrast(img, michelson_contrast, mean_luminance=None):
    """
    Adapt Michelson contrast of image

    Parameters
    ----------
    img : np.ndarray
        stimulus array
    michelson_contrast : float
        desired Michelson contrast
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    img : np.ndarray
        image with adapted michelson contrast and mean luminance if passed
    """
    if mean_luminance is None:
        mean_luminance = img.mean()

    # Adapt Michelson contrast
    img = (img - img.min()) / (img.max() - img.min())
    img = img * michelson_contrast * 2.0 * mean_luminance
    img += mean_luminance - michelson_contrast * mean_luminance
    return img


def adapt_rms_contrast(img, rms_contrast, mean_luminance=None):
    """
    Adapt rms contrast of image (std)

    Parameters
    ----------
    img : np.ndarray
        stimulus array
    rms_contrast : float
        desired rms contrast (std divided by mean intensity)
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    img : np.ndarray
        image with adapted rms contrast and mean luminance if passed
    """
    if mean_luminance is None:
        mean_luminance = img.mean()

    # Adapt rms contrast
    img = img - img.mean()
    img = img / img.std() * rms_contrast + mean_luminance
    return img


def adapt_normalized_rms_contrast(img, rms_contrast, mean_luminance=None):
    """
    Adapt normalized rms contrast of image (std divided by mean)

    Parameters
    ----------
    img : np.ndarray
        stimulus array
    rms_contrast : float
        desired rms contrast (std divided by mean intensity)
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    img : np.ndarray
        image with adapted rms contrast and mean luminance if passed
    """
    if mean_luminance is None:
        mean_luminance = img.mean()

    img = img - img.mean()
    img = img / img.std() * rms_contrast * mean_luminance + mean_luminance
    return img


def adapt_intensity_range(img, intensity_min=0.0, intensity_max=1.0):
    """
    Adapt intensity range of image

    Parameters
    ----------
    img : np.ndarray
        stimulus array
    intensity_min : float
        new minimal intensity value
    intensity_max : float
        new maximal intensity value

    Returns
    ----------
    img : np.ndarray
        image with adapted intensity range
    """

    img = (img - img.min()) / (img.max() - img.min())
    img = img * (intensity_max - intensity_min) + intensity_min
    return img


def adapt_michelson_contrast_dict(stim, michelson_contrast, mean_luminance=None):
    """
    Adapt Michelson contrast of image in dict

    Parameters
    ----------
    stim : dict
        stimulus dictionary containing at least key "img"
    michelson_contrast : float
        desired Michelson contrast
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        Michelson contrast (key: "michelson_contrast"),
        mean luminance ("mean_luminance")
        and additional keys containing stimulus parameters
    """

    # Adapt Michelson contrast of image
    img = adapt_michelson_contrast(stim["img"], michelson_contrast, mean_luminance)

    stim["img"] = img
    stim["michelson_contrast"] = michelson_contrast
    stim["mean_luminance"] = mean_luminance
    return stim


def adapt_rms_contrast_dict(stim, rms_contrast, mean_luminance=None):
    """
    Adapt rms contrast of image (std)

    Parameters
    ----------
    stim : dict
        stimulus dictionary containing at least key "img"
    rms_contrast : float
        desired rms contrast (std divided by mean intensity)
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        RMS contrast (key: "rms_contrast"),
        mean luminance ("mean_luminance")
        and additional keys containing stimulus parameters
    """
    # Adapt rms_contrast of image
    img = adapt_rms_contrast(stim["img"], rms_contrast, mean_luminance)

    stim["img"] = img
    stim["rms_contrast"] = rms_contrast
    stim["mean_luminance"] = mean_luminance
    return stim


def adapt_normalized_rms_contrast_dict(stim, rms_contrast, mean_luminance=None):
    """
    Adapt normalized rms contrast of image (std divided by mean)

    Parameters
    ----------
    stim : dict
        stimulus dictionary containing at least key "img"
    rms_contrast : float
        desired rms contrast (std divided by mean intensity)
    mean_luminance : float
        desired mean luminance; if None (default), dont change mean luminance

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        RMS contrast (key: "rms_contrast"),
        mean luminance ("mean_luminance")
        and additional keys containing stimulus parameters
    """

    # Adapt normalized rms contrast
    img = adapt_normalized_rms_contrast(stim["img"], rms_contrast, mean_luminance)

    stim["img"] = img
    stim["rms_contrast"] = rms_contrast
    stim["mean_luminance"] = mean_luminance
    return stim


def adapt_intensity_range_dict(stim, intensity_min=0.0, intensity_max=1.0):
    """
    Adapt intensity range of image

    Parameters
    ----------
    stim : dict
        stimulus dictionary containing at least key "img"
    intensity_min : float
        new minimal intensity value
    intensity_max : float
        new maximal intensity value

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        intensity range (key: "intensity_range"),
        and additional keys containing stimulus parameters
    """

    img = adapt_intensity_range(stim["img"], intensity_min, intensity_max)

    stim["img"] = img
    stim["intensity_range"] = (intensity_min, intensity_max)
    return stim



import warnings

import numpy as np

from stimupy.components.shapes import cross, rectangle, triangle
from stimupy.utils import resolution

__all__ = [
    "cross_generalized",
    "cross_rectangles",
    "cross_triangles",
    "todorovic_generalized",
    "todorovic_rectangles",
    "todorovic_triangles",
]


def cross_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_thickness=None,
    target_size=None,
    target_type="r",
    target_rotation=0.0,
    target_x=None,
    target_y=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """Benary's Cross Illusion

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_thickness : float
        width of the cross bars in degrees visual angle
    target_size : (float, float)
        size of all target(s) in degrees visual angle
    target_type : tuple of strings
        type of targets to use; option: r (rectangle), t (triangle); as many targets as types
    target_rotation : tuple of floats, or float
        tuple with rotation of targets in deg, counterclockwise, as many targets as rotations
    target_x : tuple of floats
        tuple with x coordinates of targets in degrees, as many targets as coordinates
    target_y : tuple of floats
        tuple with y coordinates of targets in degrees, as many targets as coordinates
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if cross_thickness is None:
        raise ValueError(
            "cross_generalized() missing argument 'cross_thickness' which is not 'None'"
        )

    # Create cross
    cross_stim = cross(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        cross_size=visual_size,
        cross_arm_ratios=1.0,
        cross_thickness=cross_thickness,
        intensity_background=intensity_background,
        intensity_cross=intensity_cross,
    )

    # Add targets
    stim = add_targets(
        cross_stim["img"],
        np.unique(ppd),
        target_size,
        target_type,
        target_rotation,
        target_x,
        target_y,
        intensity_target,
    )

    del cross_stim["img"]
    return {**stim, **cross_stim}


def cross_rectangles(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_thickness=None,
    target_size=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """
    Benary's Cross stimulus with two rectangular targets with default placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_thickness : float
        width of the cross bars in degrees visual angle
    target_size : (float, float)
        size of all target(s) in degrees visual angle
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if cross_thickness is None:
        raise ValueError(
            "cross_rectangles() missing argument 'cross_thickness' which is not 'None'"
        )

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if target_size is None:
        raise ValueError("target_size cannot be None")
    if isinstance(target_size, (float, int)):
        target_size = (target_size, target_size)
    if target_size[0] > cross_thickness:
        raise ValueError("Target size is larger than cross thickness")

    # Calculate target placement for classical Benarys cross
    theight, twidth = resolution.shape_from_visual_size_ppd(target_size, ppd)
    cheight, cwidth = resolution.shape_from_visual_size_ppd(cross_thickness, ppd)

    target_x = (
        (shape[1] / 2 - cwidth / 2 - twidth) / ppd[1],
        (shape[1] - twidth) / ppd[1],
    )

    target_y = (
        (shape[0] / 2 - cheight / 2 - theight) / ppd[0],
        (shape[0] / 2 - cheight / 2) / ppd[0],
    )

    stim = cross_generalized(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        cross_thickness=cross_thickness,
        target_size=target_size,
        target_type=("r",) * 2,
        target_rotation=0.0,
        target_x=np.round(np.array(target_x) * ppd) / ppd,
        target_y=np.round(np.array(target_y) * ppd) / ppd,
        intensity_background=intensity_background,
        intensity_cross=intensity_cross,
        intensity_target=intensity_target,
    )
    return stim


def cross_triangles(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_thickness=None,
    target_size=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """
    Benary's Cross stimulus with two triangular targets with default placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_thickness : float
        width of the cross bars in degrees visual angle
    target_size : float
        size of adjacent and opposite of target triangle in degrees visual angle
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
     Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if cross_thickness is None:
        raise ValueError(
            "cross_triangles() missing argument 'cross_thickness' which is not 'None'"
        )

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if target_size is None:
        raise ValueError("target_size cannot be None")
    if isinstance(target_size, (float, int)):
        target_size = (target_size, target_size)
    if target_size[0] != target_size[1]:
        raise ValueError("target needs to have the same height and width")
    if np.sqrt(target_size[0] ** 2 * 2.0) / 2.0 > cross_thickness:
        raise ValueError("Target size is larger than cross thickness")

    # Calculate target placement for classical Benarys cross
    theight, twidth = resolution.shape_from_visual_size_ppd(target_size, ppd)
    cheight, cwidth = resolution.shape_from_visual_size_ppd(cross_thickness, ppd)

    target_x = (
        (shape[1] / 2 - cwidth / 2 - twidth) / ppd[1],
        (shape[1] / 2 + cwidth / 2) / ppd[1],
    )

    target_y = (
        (shape[0] / 2 - cheight / 2 - theight) / ppd[0],
        (shape[0] / 2 - cheight / 2) / ppd[0],
    )

    stim = cross_generalized(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        cross_thickness=cross_thickness,
        target_size=target_size,
        target_type=("t",) * 2,
        target_rotation=(90.0, 45.0),
        target_x=np.round(np.array(target_x) * ppd) / ppd,
        target_y=np.round(np.array(target_y) * ppd) / ppd,
        intensity_background=intensity_background,
        intensity_cross=intensity_cross,
        intensity_target=intensity_target,
    )
    return stim


def todorovic_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    L_width=None,
    target_size=None,
    target_type="r",
    target_rotation=0.0,
    target_x=None,
    target_y=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """
    Todorovic Benary's Cross Illusion

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    L_width : float
        width of L in degree visual angle
    target_size : (float, float)
        size of all target(s) in degrees visual angle
    target_type : tuple of strings
        type of targets to use; option: r (rectangle), t (triangle); as many targets as types
    target_rotation : tuple of floats, or float
        tuple with rotation of targets in deg, counterclockwise, as many targets as rotations
    target_x : tuple of floats
        tuple with x coordinates of targets in degrees, as many targets as coordinates
    target_y : tuple of floats
        tuple with y coordinates of targets in degrees, as many targets as coordinates
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if L_width is None:
        raise ValueError("todorovic_generalized() missing argument 'L_width' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if L_width > visual_size[1] / 2:
        raise ValueError("L_width cannot be larger than stimulus_width / 2")

    L_size = (visual_size[0] / 2, visual_size[0] / 2, L_width, visual_size[1] - L_width)
    top, bottom, left, right = resolution.lengths_from_visual_angles_ppd(L_size, np.unique(ppd))
    width, height = left + right, top + bottom

    # Create stimulus without targets
    mask = np.zeros(shape).astype(int)
    mask[:, 0:left] = 1
    mask[height - bottom : :, 0 : width - left] = 1
    img = np.where(mask, intensity_cross, intensity_background)

    # Add targets
    stim = add_targets(
        img,
        np.unique(ppd),
        target_size,
        target_type,
        target_rotation,
        target_x,
        target_y,
        intensity_target,
    )

    # Add missing parameter information
    stim["visual_size"] = visual_size
    stim["ppd"] = ppd
    stim["shape"] = shape
    stim["L_width"] = L_width
    stim["intensity_background"] = intensity_background
    stim["intensity_cross"] = intensity_cross
    stim["L_mask"] = mask
    return stim


def todorovic_rectangles(
    visual_size=None,
    ppd=None,
    shape=None,
    L_width=None,
    target_size=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """
    Todorovic version of Benary's Cross stimulus with two rectangular targets
    and default placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    L_width : float
        width of L in degree visual angle
    target_size : (float, float)
        size of all target(s) in degrees visual angle
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if L_width is None:
        raise ValueError("todorovic_rectangles() missing argument 'L_width' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if target_size is None:
        raise ValueError("target_size cannot be None")
    if isinstance(target_size, (float, int)):
        target_size = (target_size, target_size)

    # Calculate target placement for classical Todorovic-Benary
    target_x = (L_width, visual_size[1] - L_width - target_size[1])
    target_y = (visual_size[0] / 2.0 - target_size[0], visual_size[0] / 2.0)

    stim = todorovic_generalized(
        visual_size=visual_size,
        ppd=ppd,
        L_width=L_width,
        target_size=target_size,
        target_type=("r",) * 2,
        target_rotation=0.0,
        target_x=np.round(np.array(target_x) * ppd) / ppd,
        target_y=np.round(np.array(target_y) * ppd) / ppd,
        intensity_background=intensity_background,
        intensity_cross=intensity_cross,
        intensity_target=intensity_target,
    )
    return stim


def todorovic_triangles(
    visual_size=None,
    ppd=None,
    shape=None,
    L_width=None,
    target_size=None,
    intensity_background=1.0,
    intensity_cross=0.0,
    intensity_target=0.5,
):
    """
    Todorovic version of Benary's Cross stimulus with two triangular targets
    and default placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    L_width : float
        width of L in degree visual angle
    target_size : float
        size of adjacent and opposite of target triangle in degrees visual angle
    intensity_background : float
        intensity value for background
    intensity_cross : float
        intensity value for cross
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
     Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """
    if L_width is None:
        raise ValueError("todorovic_triangles() missing argument 'L_width' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if target_size is None:
        raise ValueError("target_size cannot be None")
    if isinstance(target_size, (float, int)):
        target_size = (target_size, target_size)
    else:
        raise ValueError("target_size should be a single float")

    # Calculate target placement for classical Todorovic-Benary
    target_x = (L_width, visual_size[1] - L_width - target_size[1])
    target_y = (visual_size[0] / 2.0 - target_size[0], visual_size[0] / 2.0)

    stim = todorovic_generalized(
        visual_size=visual_size,
        ppd=ppd,
        L_width=L_width,
        target_size=target_size,
        target_type=("t",) * 2,
        target_rotation=(0.0, 180.0),
        target_x=np.round(np.array(target_x) * ppd) / ppd,
        target_y=np.round(np.array(target_y) * ppd) / ppd,
        intensity_background=intensity_background,
        intensity_cross=intensity_cross,
        intensity_target=intensity_target,
    )
    return stim


def add_targets(
    img,
    ppd,
    target_size,
    target_type,
    target_rotation,
    target_x,
    target_y,
    intensity_target,
):
    """Add targets to Benary-like stimulus

    Parameters
    ----------
    img : 2d numpy array
        image to which targets will be added
    ppd : int
        pixels per degree (visual angle)
    target_size : (float, float)
        size of all target(s) in degrees visual angle
    target_type : tuple of strings
        type of targets to use; option: r (rectangle), t (triangle); as many targets as types
    target_rotation : tuple of floats, or float
        tuple with rotation of targets in deg, counterclockwise, as many targets as rotations
    target_x : tuple of floats
        tuple with x coordinates of targets in degrees, as many targets as coordinates
    target_y : tuple of floats
        tuple with y coordinates of targets in degrees, as many targets as coordinates
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the updated stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """

    mask = np.zeros(img.shape)

    # If any target information is missing, dont add targets
    if (
        (target_size is None)
        or (target_type is None)
        or (target_rotation is None)
        or (target_x is None)
        or (target_y is None)
    ):
        warnings.warn("Target information is missing - no target added")
        stim = {
            "target_size": None,
            "target_type": None,
            "target_rotation": None,
            "target_x": None,
            "target_y": None,
            "intensity_target": None,
        }

    # Re-format all target information and add targets
    else:
        if isinstance(target_size, (float, int)):
            target_size = (target_size, target_size)
        if isinstance(target_type, (str)):
            target_type = [target_type]
        if isinstance(target_rotation, (float, int)):
            target_rotation = [target_rotation]
        if isinstance(target_x, (float, int)):
            target_x = [target_x]
        if isinstance(target_y, (float, int)):
            target_y = [target_y]
        if np.min(target_x) < 0:
            raise ValueError("Leftmost target does not fit into image")
        if np.min(target_y) < 0:
            raise ValueError("Topmost target does not fit into image")

        n_targets = np.max(
            np.array([len(target_type), len(target_rotation), len(target_x), len(target_y)])
        )

        if len(target_type) == 1:
            target_type = target_type * n_targets
        if len(target_rotation) == 1:
            target_rotation = target_rotation * n_targets
        if len(target_x) == 1:
            target_x = target_x * n_targets
        if len(target_y) == 1:
            target_y = target_y * n_targets

        if any(len(lst) != n_targets for lst in [target_rotation, target_x, target_y]):
            raise Exception(
                "target_type, target_rotation, target_x and target_y need the same length."
            )

        theight, twidth = resolution.lengths_from_visual_angles_ppd(target_size, ppd)
        ty = resolution.lengths_from_visual_angles_ppd(target_y, ppd)
        tx = resolution.lengths_from_visual_angles_ppd(target_x, ppd)

        if isinstance(ty, (int, float)):
            ty = (ty,)
            tx = (tx,)

        if (twidth + np.array(tx)).max() > img.shape[1]:
            raise ValueError("Rightmost target does not fit in image.")
        if (theight + np.array(ty)).max() > img.shape[0]:
            raise ValueError("Lowest target does not fit in image.")

        # Add targets:
        for i in range(len(target_x)):
            if target_type[i] == "r":
                target = rectangle(
                    shape=[theight * 2, twidth * 2],
                    ppd=ppd,
                    rectangle_size=(theight / ppd, twidth / ppd),
                    intensity_rectangle=intensity_target,
                    intensity_background=0,
                    rotation=target_rotation[i],
                )
                mpatch = target["rectangle_mask"][~np.all(target["rectangle_mask"] == 0, axis=1)]
                tpatch = target["img"][~np.all(target["rectangle_mask"] == 0, axis=1)]

            elif target_type[i] == "t":
                target = triangle(
                    shape=[theight * 3, twidth * 3],
                    ppd=ppd,
                    triangle_size=(theight / ppd, twidth / ppd),
                    intensity_background=0,
                    intensity_triangle=intensity_target,
                    rotation=target_rotation[i],
                    include_corners=True,
                )
                mpatch = target["triangle_mask"][~np.all(target["triangle_mask"] == 0, axis=1)]
                tpatch = target["img"][~np.all(target["triangle_mask"] == 0, axis=1)]

            else:
                raise Exception("You can only use r or t as shapes")

            # Remove zero-rows and -columns
            tpatch = tpatch[:, ~np.all(mpatch == 0, axis=0)]
            mpatch = mpatch[:, ~np.all(mpatch == 0, axis=0)]
            theight_, twidth_ = tpatch.shape

            if ty[i] + theight_ > img.shape[0] or tx[i] + twidth_ > img.shape[1]:
                raise ValueError("At least one target does not fully fit into stimulus")

            # Only change the target parts of the image:
            mlarge = np.zeros(img.shape)
            mlarge[ty[i] : ty[i] + theight_, tx[i] : tx[i] + twidth_] = mpatch
            tlarge = np.zeros(img.shape)
            tlarge[ty[i] : ty[i] + theight_, tx[i] : tx[i] + twidth_] = tpatch
            img = np.where(mlarge, tlarge, img)
            mask = np.where(mlarge, i + 1, mask)

        stim = {
            "target_size": target_size,
            "target_type": target_type,
            "target_rotation": target_rotation,
            "target_x": target_x,
            "target_y": target_y,
            "intensity_target": intensity_target,
        }

    stim["img"] = img
    stim["target_mask"] = mask.astype(int)
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    params_benary = {
        "target_size": 1,
        "cross_thickness": 2,
    }

    params_todo = {
        "target_size": 1,
        "L_width": 2,
    }

    target_pos = {
        "target_x": (3, 6, 3, 6),
        "target_y": (4, 6, 6, 4),
    }

    # fmt: off
    stimuli = {
        "benarys_cross_general": cross_generalized(**default_params, **params_benary, **target_pos),
        "benarys_cross_rectangles": cross_rectangles(**default_params, **params_benary),
        "benarys_cross_triangles": cross_triangles(**default_params, **params_benary),
        "benarys_todorovic_general": todorovic_generalized(**default_params, **params_todo, **target_pos),
        "benarys_todorovic_rectangles": todorovic_rectangles(**default_params, **params_todo),
        "benarys_todorovic_triangles": todorovic_triangles(**default_params, **params_todo),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import collections

import numpy as np

from stimupy.components import draw_regions, mask_regions
from stimupy.utils import resolution

__all__ = [
    "disc",
    "ring",
    "annulus",
    "rings",
]


def mask_rings(
    radii,
    visual_size=None,
    ppd=None,
    shape=None,
    origin="mean",
):
    """Generate mask with integer indices for rings

    Parameters
    ----------
    radii : Sequence[Number]
        outer radii of rings (& disc) in degree visual angle
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each ring (key: "ring_mask"),
        and additional keys containing stimulus parameters

    Raises
    ------
    ValueError
        if largest radius does not fit in visual size
    """

    # no axes are None; check if fits
    if visual_size.height < np.max(radii) * 2 or visual_size.width < np.max(radii) * 2:
        raise ValueError(
            f"Largest radius {np.max(radii)} does not fit in visual size {visual_size}"
        )

    # Mark elements with integer idx-value
    stim = mask_regions(
        distance_metric="radial",
        edges=radii,
        rotation=0.0,
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        origin=origin,
    )
    stim["ring_mask"] = stim["mask"]
    del stim["mask"]
    return stim


def rings(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    intensity_rings=(0.0, 1.0),
    intensity_background=0.5,
    origin="mean",
):
    """Draw a central solid disc with zero or more solid rings (annuli)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number]
        outer radii of rings (& disc) in degree visual angle
    intensity_rings : Sequence[Number, ...], optional
        intensity value for each ring, from inside to out, by default (0.0, 1.0)
        If fewer intensities are passed than number of radii, cycles through intensities
    intensity_background : float, optional
        value of background, by default 0.5
    origin : "corner", "mean" or "center", optional
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each ring (key: "ring_mask"),
        and additional keys containing stimulus parameters
    """
    # Try to resolve resolution;
    try:
        shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    except resolution.TooManyUnknownsError:
        # Check visual_size
        visual_size = resolution.validate_visual_size(visual_size)

    if visual_size is None or visual_size == (None, None):
        # Two axes are None; make image large enough to fit
        visual_size = resolution.validate_visual_size(np.max(radii) * 2)
    elif None in visual_size:
        # one axis is None; make square
        visual_size = [x for x in visual_size if x is not None]
        visual_size = resolution.validate_visual_size(visual_size)

    # Resolve radii
    if radii is None:
        raise ValueError("rings() missing argument 'radii' which is not 'None'")
    if isinstance(radii, collections.abc.Sequence) or isinstance(radii, np.ndarray):
        if np.diff(radii).min() < 0:
            raise ValueError("radii need to monotonically increase")
    else:
        radii = (radii,)

    # Get masks for rings
    stim = mask_rings(radii=radii, shape=shape, visual_size=visual_size, ppd=ppd, origin=origin)

    # Resolve intensities
    if not isinstance(intensity_rings, collections.abc.Sequence):
        intensity_rings = (intensity_rings,)

    # Draw rings
    stim["img"] = draw_regions(
        stim["ring_mask"], intensity_rings, intensity_background=intensity_background
    )

    # Assemble output
    stim["intensity_rings"] = intensity_rings
    stim["radii"] = radii
    stim["intensity_background"] = intensity_background
    return stim


def disc(
    visual_size=None,
    ppd=None,
    shape=None,
    radius=None,
    intensity_disc=1.0,
    intensity_background=0.0,
    origin="mean",
):
    """Draw a central disc

    Essentially, `dics(radius)` is an alias for `ring(radii=[0, radius])`

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radius : Number
        outer radius of disc in degree visual angle
    intensity_disc : Number, optional
        intensity value of disc, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    origin : "corner", "mean" or "center", optional
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each ring (key: "ring_mask"),
        and additional keys containing stimulus parameters
    """
    if radius is None:
        raise ValueError("disc() missing argument 'radius' which is not 'None'")

    if isinstance(radius, np.ndarray):
        try:
            radius = float(radius)
        except TypeError as e:
            raise ValueError("Can only pass 1 radius") from e
    elif isinstance(radius, collections.abc.Sequence):
        if len(radius) != 1:
            raise ValueError("Can only pass 1 radius")

    stim = rings(
        radii=radius,
        intensity_rings=intensity_disc,
        visual_size=visual_size,
        ppd=ppd,
        intensity_background=intensity_background,
        shape=shape,
        origin=origin,
    )
    stim["radius"] = stim.pop("radii")
    stim["intensity_disc"] = stim.pop("intensity_rings")
    return stim


def ring(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    intensity_ring=1.0,
    intensity_background=0.0,
    origin="mean",
):
    """Draw a ring (annulus)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number, Number]
        inner and outer radius of ring in degree visual angle
    intensity_ring : Number, optional
        intensity value of ring, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    origin : "corner", "mean" or "center", optional
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each ring (key: "ring_mask"),
        and additional keys containing stimulus parameters

    Raises
    ------
    ValueError
        if passed in less/more than 2 radii (inner, outer)
    ValueError
        if passed in less/more than 1 intensity
    """
    if radii is None:
        raise ValueError("ring() missing argument 'radii' which is not 'None'")
    if len(radii) != 2:
        raise ValueError("Can only pass exactly 2 radii")
    if len([intensity_ring]) != 1:
        raise ValueError("Can only pass 1 intensity")

    if radii[1] is None:
        shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
        radii[1] = np.max(visual_size) / 2

    if radii[1] < radii[0]:
        raise ValueError("first radius needs to be smaller than second radius")

    stim = rings(
        radii=radii,
        intensity_rings=[intensity_background, intensity_ring],
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        intensity_background=intensity_background,
        origin=origin,
    )
    stim["ring_mask"] = np.where(stim["ring_mask"] == 2, 1, 0)
    stim["intensity_ring"] = intensity_ring
    return stim


annulus = ring


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "radials_disc": disc(**default_params, radius=3),
        "radials_rings": rings(**default_params, radii=(1, 2, 3)),
        "radials_ring": ring(**default_params, radii=(1, 2)),
        "radials_annulus": annulus(**default_params, radii=(1, 2)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np

from stimupy.components import image_base
from stimupy.components.shapes import ellipse

__all__ = [
    "gaussian",
]


def gaussian(
    visual_size=None,
    ppd=None,
    shape=None,
    sigma=None,
    rotation=0.0,
    intensity_max=1.0,
    origin="mean",
):
    """Create a Gaussian (envelop)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    sigma : float or (float, float)
        Sigma auf Gaussian in degree visual angle (y, x)
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    intensity_max : float
        Maximal intensity value of Gaussian
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img")
        ellipse-like mask with sigma radius and integer index (key: "gaussian_mask"),
        and additional keys containing stimulus parameters
    """
    if sigma is None:
        raise ValueError("gaussian() missing argument 'sigma' which is not 'None'")
    if isinstance(sigma, (float, int)):
        sigma = (sigma, sigma)

    # Resolve resolutions and get distances
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin=origin,
    )
    xx = base["horizontal"]
    yy = base["vertical"]

    # convert rotation parameter to radians
    theta = np.deg2rad(-rotation)

    # determine a, b, c coefficients
    a = (np.cos(theta) ** 2 / (2 * sigma[1] ** 2)) + (np.sin(theta) ** 2 / (2 * sigma[0] ** 2))
    b = -(np.sin(2 * theta) / (4 * sigma[1] ** 2)) + (np.sin(2 * theta) / (4 * sigma[0] ** 2))
    c = (np.sin(theta) ** 2 / (2 * sigma[1] ** 2)) + (np.cos(theta) ** 2 / (2 * sigma[0] ** 2))

    # create Gaussian
    gaussian = np.exp(-(a * xx**2 + 2 * b * xx * yy + c * yy**2))
    gaussian = gaussian / gaussian.max() * intensity_max

    # create mask as ellipse with sigma radius
    mask = ellipse(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radius=sigma,
        rotation=rotation,
        origin=origin,
        restrict_size=False,
    )["ellipse_mask"]

    stim = {
        "img": gaussian,
        "gaussian_mask": mask.astype(int),
        "sigma": sigma,
        "rotation": rotation,
        "visual_size": base["visual_size"],
        "shape": base["shape"],
        "ppd": base["ppd"],
        "intensity_max": intensity_max,
        "origin": origin,
    }
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 10,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "gaussians_gaussian": gaussian(**default_params, sigma=(3, 1.5), rotation=70),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np

from stimupy.noises import pseudo_white_spectrum
from stimupy.utils import bandpass, resolution
from stimupy.utils.contrast_conversions import adapt_intensity_range

__all__ = [
    "narrowband",
]


def narrowband(
    visual_size=None,
    ppd=None,
    shape=None,
    center_frequency=None,
    bandwidth=None,
    intensity_range=(0, 1),
    pseudo_noise=False,
):
    """Draw narrowband noise texture

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    center_frequency : float
        noise center frequency in cpd
    bandwidth : float
        bandwidth of the noise in octaves
    intensity_range : Sequence[Number, Number]
        minimum and maximum intensity value; default: (0, 1)
    pseudo_noise : bool
        if True, generate pseudo-random noise with ideal power spectrum.

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        and additional keys containing stimulus parameters
    """
    if center_frequency is None:
        raise ValueError("narrowband() missing argument 'center_frequency' which is not 'None'")
    if bandwidth is None:
        raise ValueError("narrowband() missing argument 'bandwidth' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    bp = bandpass(
        visual_size=visual_size, ppd=ppd, center_frequency=center_frequency, bandwidth=bandwidth
    )["img"]

    if pseudo_noise:
        # Create white noise with frequency amplitude of 1 everywhere
        white_noise_fft = pseudo_white_spectrum(shape)
    else:
        # Create white noise and fft
        white_noise = np.random.rand(*shape) * 2.0 - 1.0
        white_noise_fft = np.fft.fftshift(np.fft.fft2(white_noise))

    # Filter white noise with bandpass filter
    narrow_noise_fft = white_noise_fft * bp

    # ifft
    narrow_noise = np.fft.ifft2(np.fft.ifftshift(narrow_noise_fft))
    narrow_noise = np.real(narrow_noise)

    # Adjust intensity range:
    narrow_noise = adapt_intensity_range(narrow_noise, intensity_range[0], intensity_range[1])

    stim = {
        "img": narrow_noise,
        "noise_mask": None,
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "center_frequency": center_frequency,
        "bandwidth": bandwidth,
        "pseudo_noise": pseudo_noise,
        "intensity_range": [narrow_noise.min(), narrow_noise.max()],
    }
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 20,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "narrowbands_narrowband3": narrowband(**default_params, center_frequency=3, bandwidth=1),
        "narrowbands_narrowband9": narrowband(**default_params, center_frequency=9, bandwidth=1),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



from stimupy.components import lines
from stimupy.components.shapes import disc
from stimupy.utils import make_two_sided

__all__ = ["delboeuf", "two_sided"]


def delboeuf(
    visual_size=None,
    ppd=None,
    shape=None,
    outer_radius=None,
    outer_line_width=0,
    target_radius=None,
    intensity_outer_line=0.0,
    intensity_target=0.0,
    intensity_background=1.0,
):
    """Delboeuf's (1865) stimulus

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    outer_radius : Number
        radius of outer circle
    outer_line_width : Number
        line width of outer circle in degrees visual angle
        if 0 (default), set line width to 1 px
    target_radius : Number
        radius of target circle
    intensity_outer_line : Number
        intensity value of outer circle line (default: 0)
    intensity_target : Number
        intensity value of target (default: 0)
    intensity_background : Number
        intensity value of background (default: 1)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Delboeuf, F. J. (1865).
        Note sur certaines illusions d'optique:
        Essai d'une théorie psychophysique de la maniere
        dont l'oeil apprécie les distances et les angles.
        Bulletins de l'Académie Royale des Sciences, Lettres et
        Beaux-arts de Belgique, 19, 195-216.
    """
    if outer_radius is None:
        raise ValueError("delboeuf() missing argument 'outer_radius' which is not 'None'")
    if target_radius is None:
        raise ValueError("delboeuf() missing argument 'target_radius' which is not 'None'")

    outer = lines.circle(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radius=outer_radius,
        line_width=outer_line_width,
        intensity_line=intensity_outer_line - intensity_background,
        intensity_background=0,
    )

    inner = disc(
        radius=target_radius,
        intensity_disc=intensity_target - intensity_background,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_background=0,
    )

    inner["img"] = outer["img"] + inner["img"] + intensity_background
    inner["line_mask"] = outer["line_mask"]
    inner["outer_radius"] = outer_radius
    inner["target_radius"] = target_radius
    inner["target_mask"] = inner.pop("ring_mask")
    inner["outer_line_width"] = outer["line_width"]
    inner["intensity_outer_line"] = intensity_outer_line
    inner["intensity_target"] = intensity_target
    return inner


two_sided = make_two_sided(
    delboeuf,
    two_sided_params=(
        "outer_radius",
        "outer_line_width",
        "target_radius",
        "intensity_target",
        "intensity_outer_line",
    ),
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "delboeuf": delboeuf(**default_params, target_radius=1, outer_radius=4),
        "delboeuf_2sided": two_sided(**default_params, target_radius=1, outer_radius=(2, 1.1)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import itertools

import numpy as np

from stimupy.components import draw_regions
from stimupy.stimuli import *

__all__ = [
    "mask_targets",
    "place_targets",
    "overview",
    "plot_overview",
    "benarys",
    "bullseyes",
    "checkerboards",
    "cornsweets",
    "cubes",
    "delboeufs",
    "dungeons",
    "edges",
    "gabors",
    "gratings",
    "hermanns",
    "mondrians",
    "mueller_lyers",
    "plaids",
    "ponzos",
    "rings",
    "sbcs",
    "todorovics",
    "waves",
    "wedding_cakes",
    "whites",
]


def mask_targets(element_mask, target_indices):
    """Indicate elements as targets

    Creates a new target_mask from a mask of elements (e.g., grating bars, rings, frames, etc.),
    by indexing these elements.

    Parameters
    ----------
    element_mask : numpy.ndarray
        mask with integer values for different elements / regions in a stimulus
    target_indices : Sequence[int] or int
        index or indices of elements to be designated as targets.
        Index 0 should always refer to background region.
        Indices can be negative, which results in "counting backwards"
        from the highest index in element_mask.

    Returns
    -------
    numpy.ndarray
        target mask, with integer values indicating target regions,
        in order that they appear in target_indices

    Raises
    ------
    ValueError
        if a target_idx is greater than any value in the element_mask
    """
    if target_indices is None:
        target_indices = ()
    if isinstance(target_indices, (int, float)):
        target_indices = [
            target_indices,
        ]

    target_mask = np.zeros_like(element_mask)
    for target_idx, element_idx in enumerate(target_indices):
        if element_idx < 0:
            element_idx = int(element_mask.max()) + element_idx

        if element_idx > element_mask.max():
            raise ValueError("target_idx is outside stimulus")
        target_mask = np.where(element_mask == element_idx, target_idx + 1, target_mask)

    return target_mask


def place_targets(stim, element_mask_key, target_indices, intensity_target=0.5):
    """Place targets in stimulus

    Turns image regions/elements defined by element_mask_key
    and indicated by target_indices, into targets.
    Targets are defined in a new target_mask, and drawn into image with intensity_target.

    Parameters
    ----------
    stim : dict[str, Any]
        stimulus dictionary, with at least an "img" key, and mask indicated by element_mask_key
    element_mask_key : str
        key of the mask in stim-dict indicating image "elements"/regions
    target_indices : Sequence[int] or int
        index or indices of elements to be designated as targets
    intensity_target : float, optional
        intensity value for target, by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") with targets placed,
        mask with integer index for the target (key: "target_mask")

    See also
    --------
        mask_targets, draw_regions
    """
    stim["target_mask"] = mask_targets(
        element_mask=stim[element_mask_key], target_indices=target_indices
    )

    if isinstance(intensity_target, (int, float)):
        intensity_target = [
            intensity_target,
        ]
    intensity_target = itertools.cycle(intensity_target)

    stim["img"] = np.where(
        stim["target_mask"],
        draw_regions(
            mask=stim["target_mask"], intensities=intensity_target, intensity_background=0.0
        ),
        stim["img"],
    )
    stim["target_indices"] = target_indices
    stim["intensity_target"] = intensity_target

    return stim


def overview(skip=False):
    """Generate example stimuli from this module

    Returns
    -------
    dict[str, dict]
        Dict mapping names to individual stimulus dicts
    """
    stimuli = {}
    for stimmodule_name in __all__:
        if stimmodule_name in ["overview", "plot_overview", "mask_targets", "place_targets"]:
            continue

        print(f"Generating stimuli from {stimmodule_name}")
        # Get a reference to the actual module
        stimmodule = globals()[stimmodule_name]
        try:
            stims = stimmodule.overview()

            # Accumulate
            stimuli.update(stims)
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stimuli


def plot_overview(mask=False, save=None, units="deg"):
    """Plot overview of examples in this module (and submodules)

    Parameters
    ----------
    mask : bool or str, optional
        If True, plot mask on top of stimulus image (default: False).
        If string is provided, plot this key from stimulus dictionary as mask
    save : None or str, optional
        If None (default), do not save the plot.
        If string is provided, save plot under this name.
    units : "px", "deg" (default), or str
        what units to put on the axes, by default degrees visual angle ("deg").
        If a str other than "deg"(/"degrees") or "px"(/"pix"/"pixels") is passed,
        it must be the key to a tuple in stim

    """
    from stimupy.utils import plot_stimuli

    stims = overview(skip=True)
    plot_stimuli(stims, mask=mask, units=units, save=save)


if __name__ == "__main__":
    plot_overview()



import itertools

import numpy as np

from stimupy.components import combine_masks, draw_regions
from stimupy.components.shapes import circle
from stimupy.components.shapes import ring as ring_shape
from stimupy.stimuli import mask_targets, waves

__all__ = [
    "pinwheel",
]


def pinwheel(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_segments=None,
    segment_width=None,
    rotation=0.0,
    target_indices=(),
    target_width=None,
    target_center=None,
    intensity_segments=(0.0, 1.0),
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
):
    """Pinwheel / radial White stimulus

    Parameters
    ----------
    visual_size : (float, float)
        The shape of the stimulus in degrees of visual angle. (y,x)
    ppd : int
        pixels per degree (visual angle)
    shape : Sequence[int, int], int, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        angular frequency of angular grating, in cycles per angular degree
    n_segments : int, or None (default)
        number of segments
    segment_width : Number, or None (default)
        angular width of a single segment, in degrees
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    target_width : float, or Sequence[float, ...], optional
        target width (outer - inner radius) in deg visual angle, by default 1.0
        Can specify as many target_widths as target_indices;
        if fewer widths are passed than indices, cycles through intensities
    target_center : float, or Sequence[float, ...], optional
        center (radius) in deg visual angle where each target will be placed
        within its segment, by default 1.0.
        Can specify as many centers as target_indices;
        if fewer centers are passed than indices, cycles through intensities
    intensity_segments : Sequence[float, ...]
        intensity value for each segment, by default (1.0, 0.0).
        Can specify as many intensities as n_segments;
        If fewer intensities are passed than n_segments, cycles through intensities
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Robinson, A. E., Hammon, P. S., & de Sa, V. R. (2007).
        Explaining brightness illusions
        using spatial filtering and local response normalization.
        Vision Research, 47(12), 1631-1644.
        https://doi.org/10.1016/j.visres.2007.02.017
    """

    # Draw angular grating
    stim = waves.square_angular(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        segment_width=segment_width,
        n_segments=n_segments,
        period="ignore",
        rotation=rotation,
        origin=origin,
        intensity_segments=intensity_segments,
    )

    # Mask to circular aperture
    radius = min(stim["visual_size"]) / 2
    stim["circle_mask"] = circle(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radius=radius,
        origin=origin,
    )["circle_mask"]
    stim["segment_mask"] = np.where(stim["circle_mask"], stim["segment_mask"], 0)
    stim["img"] = np.where(stim["circle_mask"], stim["img"], intensity_background)
    stim["intensity_background"] = intensity_background

    # Target segment mask
    if isinstance(target_indices, (int, float)):
        target_indices = (target_indices,)
    target_segment_mask = mask_targets(
        element_mask=stim["grating_mask"], target_indices=target_indices
    )
    stim["target_indices"] = target_indices

    # Mask ring regions
    if target_center is None:
        target_center = radius / 2
    if isinstance(target_center, (int, float)):
        target_center = (target_center,)
    stim["target_center"] = target_center
    target_center = tuple(itertools.islice(itertools.cycle(target_center), len(target_indices)))

    if target_width is None:
        raise ValueError("pinwheel() missing argument 'target_width' which is not 'None'")
    if isinstance(target_width, (int, float)):
        target_width = (target_width,)
    stim["target_width"] = target_width
    target_width = tuple(itertools.islice(itertools.cycle(target_width), len(target_indices)))

    target_ring_masks = []
    for target_idx, (center, width) in enumerate(zip(target_center, target_width)):
        # Draw ring
        inner_radius = center - (width / 2)
        outer_radius = center + (width / 2)
        if inner_radius < 0 or outer_radius > np.min(visual_size) / 2:
            raise ValueError("target does not fully fit into pinwheel")
        ring = ring_shape(
            radii=[inner_radius, outer_radius],
            intensity_ring=target_idx,
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            shape=stim["shape"],
        )
        target_ring_masks.append(ring["ring_mask"])

    # Combine segment & ring masks
    target_masks = []
    for target_idx, ring_mask in enumerate(target_ring_masks):
        # Find where ring intesects with target segment
        target_mask = (target_segment_mask == target_idx + 1) & ring_mask
        target_masks.append(target_mask)

    # Combine target masks
    if len(target_masks) > 0:
        target_mask = combine_masks(*target_masks)
    else:
        target_mask = np.zeros_like(stim["img"])
    stim["target_mask"] = target_mask.astype(int)

    # Draw target(s)
    stim["img"] = np.where(
        target_mask, draw_regions(mask=target_mask, intensities=intensity_target), stim["img"]
    )
    stim["intensity_target"] = intensity_target

    return stim


def overview(**kwargs):
    """Generate example stimuli from this module
    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "pinwheel": pinwheel(**default_params, n_segments=10, target_width=2, target_indices=3),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np

from stimupy.utils import resolution

__all__ = [
    "grid",
]


def grid(
    visual_size=None,
    ppd=None,
    shape=None,
    element_size=None,
    intensity_background=0.0,
    intensity_grid=1.0,
):
    """Hermann's (1870) grid

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    element_size : (float, float, float)
        height, width and thickness of individual elements in degree visual angle
    intensity_background : float
        value of background
    intensity_grid : float
        value of grid

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        empty mask (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Hermann L (1870).
        Eine Erscheinung simultanen Contrastes".
        Pflügers Archiv für die gesamte Physiologie. 3: 13-15.
        https://doi.org/10.1007/BF01855743
    """
    if element_size is None:
        raise ValueError("grid() missing argument 'element_size' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    eheight, ewidth, ethick = resolution.lengths_from_visual_angles_ppd(
        element_size, np.unique(ppd)
    )

    if eheight <= ethick:
        raise ValueError("Element thickness larger than height")
    if ewidth <= ethick:
        raise ValueError("Element thickness larger than width")
    if ethick <= 0:
        raise ValueError("Increase element thickness")

    img = np.ones(shape) * intensity_background
    for i in range(ethick):
        img[i::eheight, :] = intensity_grid
        img[:, i::ewidth] = intensity_grid

    stim = {
        "img": img,
        "target_mask": np.zeros(shape).astype(int),  # TODO: add mask
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "element_size": element_size,
        "intensity_background": intensity_background,
        "intensity_grid": intensity_grid,
    }
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "hermann_grid": grid(**default_params, element_size=(1.5, 1.5, 0.2))
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import copy

import numpy as np

from stimupy.components import gaussians, image_base

__all__ = [
    "step",
    "gaussian",
    "cornsweet",
]


def step(
    visual_size=None,
    ppd=None,
    shape=None,
    rotation=0.0,
    intensity_edges=(0.0, 1.0),
):
    """Draw a central step edge

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    intensity_edges : (float, float)
        intensity values of edges (default: (0., 1.))

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the each lobes (key: "edge_mask"),
        and additional keys containing stimulus parameters
    """
    # Resolve resolutions and get distances
    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin="corner",
    )

    img = np.ones(shape) * intensity_edges[0]
    img = np.where(base["oblique"] < base["oblique"].mean(), img, intensity_edges[1])
    mask = np.ones(shape)
    mask = np.where(base["oblique"] < base["oblique"].mean(), mask, 2)

    stim = {
        "img": img,
        "edge_mask": mask.astype(int),
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "shape": base["shape"],
        "rotation": rotation,
        "intensity_edges": intensity_edges,
    }

    return stim


def gaussian(
    visual_size=None,
    ppd=None,
    shape=None,
    sigma=None,
    rotation=0.0,
    intensity_edges=(0.0, 1.0),
    intensity_background=0.5,
):
    """Draw a central step edge with a Gaussian envelop

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    sigma : float or (float, float)
        sigma of Gaussian in degree visual angle (y, x)
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    intensity_edges : (float, float)
        intensity values of edges (default: (0., 1.))
    intensity_background : float
        intensity value of background, by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the each lobes (key: "edge_mask"),
        and additional keys containing stimulus parameters
    """
    if sigma is None:
        raise ValueError("gaussian_edge() missing argument 'sigma' which is not 'None'")

    stim = step(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        intensity_edges=intensity_edges,
    )

    window = gaussians.gaussian(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        sigma=sigma,
    )

    img = stim["img"] - intensity_background
    img = img * window["img"] + intensity_background
    stim["img"] = img
    stim["sigma"] = sigma
    stim["intensity_background"] = intensity_background
    stim["gaussian_mask"] = window["gaussian_mask"]
    return stim


def cornsweet(
    visual_size=None,
    ppd=None,
    shape=None,
    ramp_width=None,
    rotation=0.0,
    intensity_edges=(0.0, 1.0),
    intensity_plateau=0.5,
    exponent=2.75,
):
    """Draw rectangular Cornsweet edge stimulus.
    The 2D luminance profile of the stimulus is defined as
    Left side:
    v = vtarget + (1 - X / w) ** a * (intensity_max-vtarget) for the ramp and v = vtarget beyond.
    Right side:
    v = vtarget - (1 - X / w) ** a * (intensity_min-vtarget) for the ramp and v = vtarget beyond.
    X is the distance to the edge, w is the width of the ramp, a is a variable
    determining the steepness of the ramp, vtarget is the luminance of the targets and
    intensity_max/intensity_min are the max/min luminances.

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    ramp_width : float
        width of luminance ramp in degrees of visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0
    intensity_edges : (float, float)
        intensity of edges
    intensity_plateau : float
        intensity value of plateau
    exponent : float
        determines steepness of ramp (default is 2.75. 1 would be linear)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the each lobes (key: "edge_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Boyaci, H., Fang, F., Murray, S.O., Kersten, D. (2007).
        Responses to lightness variations in early human visual cortex.
        Current Biology 17, 989-993.
        https://doi.org/10.1016/j.cub.2007.05.005
    Cornsweet, T. (1970).
        Visual perception. Academic press.
        https://doi.org/10.1016/B978-0-12-189750-5.X5001-5
    """
    if ramp_width is None:
        raise ValueError("cornsweet_edge() missing argument 'ramp_width' which is not 'None'")

    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin="mean",
    )

    if ramp_width > max(base["visual_size"]) / 2:
        raise ValueError("ramp_width is too large")

    dist = np.round(base["oblique"] / ramp_width, 6)
    d1 = copy.deepcopy(dist)
    d2 = copy.deepcopy(dist) * (-1)
    d1[d1 < 0] = -1
    d1[d1 > 1] = 1
    d2[d2 < 0] = -1
    d2[d2 > 1] = 1

    # Create ramp profiles individually for left and right side
    profile1 = (1.0 - d1) ** exponent * (
        intensity_edges[0] - intensity_plateau
    ) + intensity_plateau
    profile2 = (1.0 - d2) ** exponent * (
        intensity_edges[1] - intensity_plateau
    ) + intensity_plateau
    img = np.where(d1 == -1, 0, profile1) + np.where(d2 == -1, 0, profile2)
    mask = np.where(d1 == -1, 0, 2) + np.where(d2 == -1, 0, 1)
    mask[mask == 3] = 1

    stim = {
        "img": img,
        "edge_mask": mask.astype(int),
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "shape": base["shape"],
        "intensity_edges": intensity_edges,
        "intensity_plateau": intensity_plateau,
        "ramp_width": ramp_width,
        "rotation": rotation,
        "exponent": exponent,
        "d1": d1,
        "d2": d2,
    }

    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 10,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "edges_step": step(**default_params),
        "edges_gaussian": gaussian(**default_params, sigma=3),
        "edges_cornsweet": cornsweet(**default_params, ramp_width=3),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)
